package meerkat.voting.controller;

import meerkat.protobuf.Crypto;
import meerkat.protobuf.Voting.*;
import meerkat.voting.controller.callbacks.*;
import meerkat.voting.controller.commands.*;
import meerkat.voting.controller.selector.QuestionSelector;
import meerkat.voting.controller.selector.SimpleListCategoriesSelector;
import meerkat.voting.encryptor.VBCryptoManager;
import meerkat.voting.encryptor.VBCryptoManager.EncryptionAndSecrets;
import meerkat.voting.output.BallotOutputDevice;
import meerkat.voting.storage.StorageManager;
import meerkat.voting.ui.VotingBoothUI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.security.SignatureException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Created by hai on 28/03/16.
 */
public class VotingBoothImpl implements VotingBoothController {

    private BallotOutputDevice outputDevice;
    private VBCryptoManager crypto;
    private VotingBoothUI ui;
    private StorageManager storageManager;
    private List<BallotQuestion> questionsForChoosingChannel;
    private List<BallotQuestion> questions;
    private QuestionSelector questionSelector;
    private Map<String, UIElement> systemMessages;

    private LinkedBlockingQueue<ControllerCommand> queue;

    private Logger logger;

    private ControllerState state;
    private volatile boolean shutDownHasBeenCalled;

    protected final int MAX_REQUEST_IDENTIFIER = 100000;
    private static int requestCounter = 0;



    public VotingBoothImpl () {
        logger = LoggerFactory.getLogger(VotingBoothImpl.class);
        logger.info("A VotingBoothImpl is constructed");
        shutDownHasBeenCalled = false;
        queue = new LinkedBlockingQueue<>();
        state = new ControllerState();
    }

    @Override
    public void init(BallotOutputDevice outputDevice,
                     VBCryptoManager vbCrypto,
                     VotingBoothUI vbUI,
                     StorageManager vbStorageManager) throws IOException {
        logger.info("init is called");
        this.outputDevice = outputDevice;
        this.crypto = vbCrypto;
        this.ui = vbUI;
        this.storageManager = vbStorageManager;

        ElectionParams electionParams;
        try {
            logger.info("init: reading election params");
            electionParams = storageManager.readElectionParams();
            logger.info("init: reading system messages");
            systemMessages = storageManager.readSystemMessages();
        }
        catch (IOException e) {
            logger.error("init could not read info from a file. Exception is: " + e);
            throw e;
        }

        logger.info("init: setting the election parameters");
        this.questionsForChoosingChannel = electionParams.getChannelChoiceQuestionsList();
        this.questions = electionParams.getRaceQuestionsList();
        this.questionSelector = new SimpleListCategoriesSelector(this.questions, electionParams.getSelectionData());
        logger.info("init: setting finished");
    }


    @Override
    public void run() {
        logger.info("run command has been called");
        runVotingFlow();
        doShutDown();

    }

    private void runVotingFlow () {
        logger.info("entered the voting flow");

        queue.add(new RestartVotingCommand(generateRequestIdentifier(), state.currentBallotSerialNumber));

        while (! wasShutDownCalled()) {
            try {
                ControllerCommand task = queue.take();
                handleSingleTask (task);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from task queue " + e);
            }
        }
    }

    @Override
    public void callShutDown() {
        logger.info("callShutDown command has been called");
        shutDownHasBeenCalled = true;
        queue.clear();
        ui.callShutDown();
        outputDevice.callShutDown();
    }

    private void handleSingleTask (ControllerCommand task) {
        if (task.getBallotSerialNumber() != state.currentBallotSerialNumber && !(task instanceof RestartVotingCommand)) {
            // probably an old command relating to some old ballot serial number. Simply log it and ignore it.
            String errorMessage = "handleSingleTask: received a task too old. " +
                    task.getBallotSerialNumber() + " " + state.currentBallotSerialNumber;
            logger.debug(errorMessage);
            return;
        }

        if (task instanceof RestartVotingCommand) {
            doRestartVoting ();
        }
        else if (task instanceof ChannelChoiceCommand) {
            doChooseChannel();
        }
        else if (task instanceof ChannelDeterminedCommand) {
            doSetChannelAndAskQuestions ((ChannelDeterminedCommand)task);
        }
        else if (task instanceof ChooseFinalizeOptionCommand) {
            doChooseFinalizeOption();
        }
        else if (task instanceof CastCommand) {
            doFinalize(false);
        }
        else if (task instanceof AuditCommand) {
            doFinalize(true);
        }
        else if (task instanceof EncryptAndCommitBallotCommand) {
            doCommit ((EncryptAndCommitBallotCommand)task);
        }
        else if (task instanceof ReportErrorCommand) {
            doReportErrorAndForceRestart((ReportErrorCommand)task);
        }
        else {
            logger.error("handleSingleTask: unknown type of ControllerCommand received: " + task.getClass().getName());
            doReportErrorAndForceRestart(systemMessages.get(storageManager.SOMETHING_WRONG_MESSAGE));
        }
    }

    private boolean wasShutDownCalled () {
        return shutDownHasBeenCalled;
    }

    private void doShutDown () {
        logger.info("running callShutDown");
        state.clearAndResetState(VBState.SHUT_DOWN);
        //TODO: add commands to actually shut down the machine
    }

    private void doRestartVoting () {
        queue.clear();
        state.clearAndResetState(VBState.NEW_VOTER);

        ui.startNewVoterSession(new NewVoterCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
    }

    private void doReportErrorAndForceRestart(ReportErrorCommand task) {
        doReportErrorAndForceRestart(task.getErrorMessage());
    }

    private void doReportErrorAndForceRestart(UIElement errorMessage) {
        queue.clear();
        state.clearAndResetState(VBState.FATAL_ERROR_FORCE_NEW_VOTER);
        ui.showErrorMessageWithButtons(errorMessage,
                new UIElement[]{systemMessages.get(storageManager.RESTART_VOTING_BUTTON)},
                new ErrorMessageRestartCallback(generateRequestIdentifier(),
                        state.currentBallotSerialNumber,
                        this.queue));
    }

    private void doChooseChannel () {
        if (state.stateIdentifier == VBState.NEW_VOTER) {
            logger.debug("doing chooseChannel");
            state.stateIdentifier = VBState.CHOOSE_CHANNEL;
            ui.chooseChannel(this.questionsForChoosingChannel,
                    new ChannelChoiceCallback(generateRequestIdentifier(),
                            state.currentBallotSerialNumber,
                            this.queue,
                            systemMessages.get(storageManager.UNSUCCESSFUL_CHANNEL_CHOICE_MESSAGE)));
        }
        else {
            logger.debug("doChooseChannel: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }

    private void doSetChannelAndAskQuestions (ChannelDeterminedCommand task) {
        if (state.stateIdentifier == VBState.CHOOSE_CHANNEL) {
            logger.debug("doing set channel and ask questions");
            state.stateIdentifier = VBState.ANSWER_QUESTIONS;
            List<BallotAnswer> channelChoiceAnswers = task.channelChoiceAnswers;
            state.channelIdentifier = questionSelector.getChannelIdentifier(channelChoiceAnswers);
            state.channelSpecificQuestions = questionSelector.selectQuestionsForVoter(state.channelIdentifier);
            ui.askVoterQuestions(state.channelSpecificQuestions,
                    new VotingCallback(generateRequestIdentifier(),
                            state.currentBallotSerialNumber,
                            this.queue,
                            systemMessages.get(storageManager.UNSUCCESSFUL_VOTING_MESSAGE)));
        }
        else {
            logger.debug("doSetChannelAndAskQuestions: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }


    private void doChooseFinalizeOption() {
        if (state.stateIdentifier == VBState.COMMITTING_TO_BALLOT) {
            logger.debug("doChooseFinalizeOption");
            state.stateIdentifier = VBState.CAST_OR_AUDIT;
            ui.castOrAudit(new CastOrAuditCallback(generateRequestIdentifier(),
                    state.currentBallotSerialNumber,
                    this.queue,
                    systemMessages.get(storageManager.UNRECOGNIZED_FINALIZE_RESPONSE_MESSAGE)));
        }
        else {
            logger.debug("doChooseFinalizeOption: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }
    private void doCommit (EncryptAndCommitBallotCommand task) {
        if (state.stateIdentifier == VBState.ANSWER_QUESTIONS) {
            logger.debug("doing commit");
            try {
                setBallotData(task);
                ui.notifyVoterToWaitForFinish(systemMessages.get(storageManager.WAIT_FOR_COMMIT_MESSAGE),
                        new WaitForFinishCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.SOMETHING_WRONG_MESSAGE)));
                outputDevice.commitToBallot(state.plaintextBallot,
                        state.signedEncryptedBallot,
                        new OutputDeviceCommitCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.OUTPUT_DEVICE_FAILURE_MESSAGE)));
                state.stateIdentifier = VBState.COMMITTING_TO_BALLOT;
            }
            catch (SignatureException | IOException e) {
                logger.error("doCommit: encryption failed. exception: " + e);
                UIElement errorMessage = systemMessages.get(storageManager.ENCRYPTION_FAILED_MESSAGE);
                UIElement[] buttons = new UIElement[]{
                        systemMessages.get(storageManager.RETRY_BUTTON),
                        systemMessages.get(storageManager.CANCEL_VOTE_BUTTON)};

                EncryptionFailedCallback callback = new EncryptionFailedCallback(generateRequestIdentifier(),
                        state.currentBallotSerialNumber,
                        this.queue);
                ui.showErrorMessageWithButtons(errorMessage, buttons, callback);
            }
        }
        else {
            logger.debug("doCommit: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }

    private void setBallotData (EncryptAndCommitBallotCommand task) throws IOException, SignatureException{
        if (! (task instanceof RetryEncryptAndCommitBallotCommand)) {
            // this is not a retry attempt, so the plaintext is not set yet
            // otherwise, we have the plaintext from the previous encryption attempt
            state.plaintextBallot = PlaintextBallot.newBuilder()
                    .setSerialNumber(task.getBallotSerialNumber())
                    .addAllAnswers(task.getVotingAnswers())
                    .build();
        }
        EncryptionAndSecrets encryptionAndSecrets = crypto.encrypt(state.plaintextBallot);
        state.signedEncryptedBallot = encryptionAndSecrets.getSignedEncryptedBallot();
        state.secrets = encryptionAndSecrets.getSecrets();
    }

    private void doFinalize (boolean auditRequested) {
        if (state.stateIdentifier == VBState.CAST_OR_AUDIT) {
            logger.debug("finalizing");
            state.stateIdentifier = VBState.FINALIZING;
            if (auditRequested) {
                ui.notifyVoterToWaitForFinish(systemMessages.get(storageManager.WAIT_FOR_AUDIT_MESSAGE),
                        new WaitForFinishCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.SOMETHING_WRONG_MESSAGE)));
                outputDevice.audit(state.secrets,
                        new OutputDeviceFinalizeCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.OUTPUT_DEVICE_FAILURE_MESSAGE)));
            }
            else {
                ui.notifyVoterToWaitForFinish(systemMessages.get(storageManager.WAIT_FOR_CAST_MESSAGE),
                        new WaitForFinishCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.SOMETHING_WRONG_MESSAGE)));
                outputDevice.castBallot(
                        new OutputDeviceFinalizeCallback(generateRequestIdentifier(),
                                state.currentBallotSerialNumber,
                                this.queue,
                                systemMessages.get(storageManager.OUTPUT_DEVICE_FAILURE_MESSAGE)));
            }
        }
        else {
            logger.debug("doFinalize: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }





    private enum VBState {
        NEW_VOTER,
        CHOOSE_CHANNEL,
        ANSWER_QUESTIONS,
        COMMITTING_TO_BALLOT,
        CAST_OR_AUDIT,
        FINALIZING,
        FATAL_ERROR_FORCE_NEW_VOTER,
        SHUT_DOWN
    }


    private class ControllerState {
        public VBState stateIdentifier;
        public byte[] channelIdentifier;
        public List<BallotQuestion> channelSpecificQuestions;
        public PlaintextBallot plaintextBallot;
        public SignedEncryptedBallot signedEncryptedBallot;
        public BallotSecrets secrets;
        public int lastRequestIdentifier;
        public long currentBallotSerialNumber;

        public ControllerState () {
            plaintextBallot = null;
            signedEncryptedBallot = null;
            secrets = null;
            lastRequestIdentifier = -1;
            channelIdentifier = null;
            channelSpecificQuestions = null;
            currentBallotSerialNumber = 0;
        }


        private void clearPlaintext () {
            plaintextBallot = null;
        }

        private void clearCiphertext () {
            signedEncryptedBallot = null;
            secrets = null;
        }

        public void clearAndResetState(VBState newStateIdentifier) {
            state.clearPlaintext();
            state.clearCiphertext();
            state.stateIdentifier = newStateIdentifier;
            state.currentBallotSerialNumber += 1;
        }

    }


    private int generateRequestIdentifier() {
        ++requestCounter;
        if (requestCounter >= MAX_REQUEST_IDENTIFIER) {
            requestCounter = 1;
        }
        return requestCounter;
    }


}

